import { faker } from '@faker-js/faker';
import { dereferenceType, isValid, arrayType, STRING_TYPE } from '@previewjs/type-analyzer';
import assertNever from 'assert-never';
import prettier from 'prettier';
import parserBabel from 'prettier/parser-babel.js';
import ts from 'typescript';

function formatExpression(expressionSource) {
  try {
    const formattedStatement = prettier.format(`value = ${expressionSource}`, {
      parser: "babel",
      plugins: [parserBabel],
      filepath: "component.js",
      trailingComma: "none"
    }).trim();
    if (formattedStatement.startsWith("value = (\n")) {
      return stripFormattedIndent(
        formattedStatement.replace(/^value = \(((.|\s)*)\);$/m, "$1")
      );
    }
    return stripFormattedIndent(
      formattedStatement.replace(/^value = ((.|\s)*);$/m, "$1")
    );
  } catch {
    return expressionSource;
  }
}
function stripFormattedIndent(source) {
  const lines = source.split("\n");
  if (lines.length <= 1) {
    return source.trim();
  }
  const firstLine = lines[0]?.trim() ? lines[0] : lines[1];
  const firstLineIndentMatch = firstLine?.match(/^( +)/);
  if (!firstLineIndentMatch) {
    return source.trim();
  }
  const indent = firstLineIndentMatch[1]?.length || 0;
  return lines.map((line) => {
    const beginning = line.substring(0, indent);
    if (beginning.trim().length) {
      return line;
    }
    return line.substring(indent);
  }).join("\n").trim();
}

const RESERVED_IDENTIFIERS = /* @__PURE__ */ new Set(["arguments", "get", "set"]);
function isValidPropName(name) {
  return /^[a-zA-Z_$][a-zA-Z_$0-9]*$/.test(name) && !RESERVED_IDENTIFIERS.has(name);
}

function array(items) {
  return {
    kind: "array",
    items
  };
}
const EMPTY_ARRAY = array([]);
function boolean(value) {
  return {
    kind: "boolean",
    value
  };
}
const FALSE = boolean(false);
const TRUE = boolean(true);
function fn(source) {
  return {
    kind: "function",
    source
  };
}
function map(values) {
  return {
    kind: "map",
    values
  };
}
function node(tag, props, children = null) {
  return {
    kind: "node",
    tag,
    props,
    children
  };
}
const NULL = { kind: "null" };
function number(value) {
  return {
    kind: "number",
    value
  };
}
function object(entries) {
  return {
    kind: "object",
    entries
  };
}
function promise(value) {
  return {
    kind: "promise",
    value
  };
}
const EMPTY_OBJECT = object([]);
const EMPTY_MAP = map(EMPTY_OBJECT);
function set(values) {
  return {
    kind: "set",
    values
  };
}
const EMPTY_SET = set(EMPTY_ARRAY);
function string(value) {
  return {
    kind: "string",
    value
  };
}
const UNDEFINED = { kind: "undefined" };
function unknown(source) {
  return {
    kind: "unknown",
    source
  };
}
const UNKNOWN = unknown();

function serializableValueToJavaScript(value) {
  return formatExpression(serializableValueToUnformattedJavaScript(value));
}
function serializableValueToUnformattedJavaScript(value) {
  switch (value.kind) {
    case "array":
      return `[${value.items.map((item) => serializableValueToUnformattedJavaScript(item)).join(", ")}]`;
    case "boolean":
      return value.value ? "true" : "false";
    case "function":
      return value.source;
    case "map":
      return `new Map(${value.values.entries.length > 0 ? `Object.entries(${serializableValueToUnformattedJavaScript(
        value.values
      )})` : ""})`;
    case "node":
      return value.children ? `<${value.tag} ${unformattedJsxProps(value.props)}>${value.children.map((child) => {
        if (child.kind === "string" && // Whitespaces aren't safe to inline.
        child.value.trim() === child.value) {
          return child.value;
        } else if (child.kind === "node") {
          return serializableValueToJavaScript(child);
        } else {
          return `{${serializableValueToJavaScript(child)}}`;
        }
      }).join("\n")}</${value.tag}>` : `<${value.tag} ${unformattedJsxProps(value.props)} />`;
    case "null":
      return "null";
    case "number":
      return value.value.toString(10);
    case "object": {
      if (Object.entries(value.entries).length === 0) {
        return "{}";
      }
      let text = "";
      text += "{\n";
      for (const entry of value.entries) {
        if (entry.kind === "key") {
          text += `${entry.key.kind === "string" ? JSON.stringify(entry.key.value) : `[${serializableValueToUnformattedJavaScript(entry.key)}]`}: ${serializableValueToUnformattedJavaScript(entry.value)},
`;
        } else if (entry.kind === "spread") {
          text += `...${serializableValueToJavaScript(entry.value)},
`;
        } else {
          throw assertNever(entry);
        }
      }
      text += "\n}";
      return text;
    }
    case "promise":
      if (value.value.type === "reject") {
        return `Promise.reject(${value.value.message === null ? "" : `new Error(${JSON.stringify(value.value.message)})`})`;
      } else {
        return `Promise.resolve(${serializableValueToUnformattedJavaScript(
          value.value.value
        )})`;
      }
    case "set":
      return `new Set(${value.values.items.length > 0 ? serializableValueToUnformattedJavaScript(value.values) : ""})`;
    case "string":
      return JSON.stringify(value.value);
    case "undefined":
      return "undefined";
    case "unknown":
      return value.source ?? "{}";
    default:
      throw assertNever(value);
  }
}
function unformattedJsxProps(props) {
  const attributes = [];
  for (const prop of props.entries) {
    if (prop.kind === "spread") {
      attributes.push(`{...(${serializableValueToJavaScript(prop.value)})}`);
    } else if (prop.key.kind !== "string") {
      attributes.push(`...(${serializableValueToJavaScript(object([prop]))})`);
    } else if (prop.value.kind === "boolean" && prop.value.value === true) {
      attributes.push(`${prop.key.value}`);
    } else if (prop.value.kind === "string") {
      attributes.push(`${prop.key.value}=${JSON.stringify(prop.value.value)}`);
    } else {
      attributes.push(
        `${prop.key.value}={${serializableValueToJavaScript(prop.value)}}`
      );
    }
  }
  return attributes.join("\n");
}

function generateSerializableValue(type, collected, options = {}) {
  return _generateSerializableValue(
    type,
    collected,
    options.fieldName || "",
    [],
    options.random || false,
    false
  );
}
function _generateSerializableValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  let encounteredAliases;
  [type, encounteredAliases] = dereferenceType(
    type,
    collected,
    rejectTypeNames
  );
  rejectTypeNames = [...rejectTypeNames, ...encounteredAliases];
  switch (type.kind) {
    case "any":
    case "unknown":
    case "never":
    case "void":
      return UNDEFINED;
    case "null":
      return NULL;
    case "boolean":
      return random && Math.random() < 0.5 ? TRUE : FALSE;
    case "string":
      return string(
        random ? faker.lorem.words(generateRandomInteger(0, 10)) : fieldName
      );
    case "node":
      return node("div", EMPTY_OBJECT, [
        _generateSerializableValue(
          STRING_TYPE,
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        )
      ]);
    case "number":
      return number(random ? generateRandomInteger() : 0);
    case "literal":
      if (typeof type.value === "number") {
        return number(type.value);
      } else if (typeof type.value === "string") {
        return string(type.value);
      } else {
        return type.value ? TRUE : FALSE;
      }
    case "enum": {
      const optionValues = Object.values(type.options);
      const value = optionValues[random ? generateRandomInteger(0, optionValues.length) : 0];
      if (typeof value === "number") {
        return number(value);
      } else {
        return string(value || "unknown");
      }
    }
    case "array":
      return generateArrayValue(
        type,
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "set":
      return set(
        generateArrayValue(
          arrayType(type.items),
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        )
      );
    case "tuple":
      return array(
        type.items.map(
          (item) => _generateSerializableValue(
            item,
            collected,
            fieldName,
            rejectTypeNames,
            random,
            isFunctionReturnValue
          )
        )
      );
    case "object": {
      const entries = [];
      for (const [propName, propType] of Object.entries(type.fields)) {
        let nonOptionalPropType = propType.kind === "optional" ? propType.type : propType;
        if (propType.kind === "optional" && (!random || Math.random() < 0.5)) {
          continue;
        }
        const propValue = _generateSerializableValue(
          nonOptionalPropType,
          collected,
          propName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        );
        if (propValue.kind === "undefined") {
          continue;
        }
        if (!isValidPropName(propName)) {
          continue;
        }
        entries.push({
          kind: "key",
          key: string(propName),
          value: propValue
        });
      }
      return object(entries);
    }
    case "map":
      return map(
        generateRecordValue(
          {
            ...type,
            kind: "record"
          },
          collected,
          fieldName,
          rejectTypeNames,
          random,
          isFunctionReturnValue
        )
      );
    case "record":
      return generateRecordValue(
        {
          ...type,
          kind: "record"
        },
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "union":
      if (!random) {
        if (isValid(type, collected, void 0)) {
          return UNDEFINED;
        }
        if (isValid(type, collected, null)) {
          return NULL;
        }
        if (isValid(type, collected, false)) {
          return FALSE;
        }
      }
      return _generateSerializableValue(
        type.types[random ? generateRandomInteger(0, type.types.length) : 0],
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "intersection":
      return _generateSerializableValue(
        type.types[0],
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    case "function": {
      if (isFunctionReturnValue) {
        return fn(`() => {}`);
      }
      const returnValue = serializableValueToJavaScript(
        _generateSerializableValue(
          type.returnType,
          collected,
          fieldName,
          rejectTypeNames,
          random,
          true
        )
      );
      return fn(
        formatExpression(
          `() => {
            console.log(${JSON.stringify(fieldName + " invoked")});
            ${returnValue === "undefined" ? "" : `return ${returnValue};`}
          }`
        )
      );
    }
    case "promise": {
      return promise({
        type: "reject",
        message: null
      });
    }
    case "name":
      return _generateSerializableValue(
        type,
        collected,
        fieldName,
        rejectTypeNames,
        random,
        isFunctionReturnValue
      );
    default:
      throw assertNever(type);
  }
}
function generateArrayValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  if (isFunctionReturnValue) {
    return EMPTY_ARRAY;
  }
  const itemValues = [];
  const length = random ? generateRandomInteger(0, 3) : 1;
  for (let i = 0; i < length; i++) {
    const itemValue = _generateSerializableValue(
      type.items,
      collected,
      fieldName,
      rejectTypeNames,
      random,
      isFunctionReturnValue
    );
    itemValues.push(itemValue);
  }
  if (itemValues.every(
    (itemValue) => itemValue.kind === "undefined" || itemValue.kind === "object" && Object.keys(itemValue.entries).length === 0
  )) {
    return EMPTY_ARRAY;
  }
  return array(itemValues);
}
function generateRecordValue(type, collected, fieldName, rejectTypeNames, random, isFunctionReturnValue) {
  if (!random) {
    return EMPTY_OBJECT;
  }
  const values = generateArrayValue(
    {
      kind: "array",
      items: type.values
    },
    collected,
    fieldName,
    rejectTypeNames,
    random,
    isFunctionReturnValue
  ).items;
  const entries = [];
  for (const value of values) {
    const key = _generateSerializableValue(
      type.keys,
      collected,
      fieldName,
      rejectTypeNames,
      random,
      isFunctionReturnValue
    );
    entries.push({
      kind: "key",
      key,
      value
    });
  }
  return object(entries);
}
function generateRandomInteger(minInclusive = -5e3, maxExclusive = 5e3) {
  const range = maxExclusive - minInclusive;
  return Math.floor(Math.random() * range + minInclusive);
}

function parseSerializableValue(expression) {
  const fallbackValue = unknown(expression.getText());
  if (ts.isParenthesizedExpression(expression)) {
    return parseSerializableValue(expression.expression);
  }
  if (expression.kind === ts.SyntaxKind.NullKeyword) {
    return NULL;
  }
  if (expression.kind === ts.SyntaxKind.TrueKeyword) {
    return TRUE;
  }
  if (expression.kind === ts.SyntaxKind.FalseKeyword) {
    return FALSE;
  }
  if (expression.kind === ts.SyntaxKind.UndefinedKeyword || ts.isIdentifier(expression) && expression.text === "undefined") {
    return UNDEFINED;
  }
  if (ts.isNumericLiteral(expression)) {
    try {
      return number(parseFloat(expression.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts.isPrefixUnaryExpression(expression) && expression.operator === ts.SyntaxKind.MinusToken && ts.isNumericLiteral(expression.operand)) {
    try {
      return number(-parseFloat(expression.operand.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts.isStringLiteral(expression)) {
    return string(expression.text);
  }
  if (ts.isNoSubstitutionTemplateLiteral(expression)) {
    return string(expression.text);
  }
  if (ts.isArrayLiteralExpression(expression)) {
    return array(expression.elements.map(parseSerializableValue));
  }
  if (ts.isNewExpression(expression) && (expression.arguments?.length || 0) <= 1) {
    if (ts.isIdentifier(expression.expression)) {
      if (expression.expression.text === "Set") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_SET;
        } else if (ts.isArrayLiteralExpression(firstArgument)) {
          return set(array(firstArgument.elements.map(parseSerializableValue)));
        }
      }
      if (expression.expression.text === "Map") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_MAP;
        } else if (ts.isArrayLiteralExpression(firstArgument)) {
          const entries = [];
          for (const element of firstArgument.elements) {
            if (!ts.isArrayLiteralExpression(element) || element.elements.length !== 2) {
              return fallbackValue;
            }
            entries.push({
              kind: "key",
              key: parseSerializableValue(element.elements[0]),
              value: parseSerializableValue(element.elements[1])
            });
          }
          return map(object(entries));
        } else if (ts.isCallExpression(firstArgument) && ts.isPropertyAccessExpression(firstArgument.expression) && ts.isIdentifier(firstArgument.expression.expression) && firstArgument.expression.expression.text === "Object" && firstArgument.expression.name.text === "entries" && firstArgument.arguments.length === 1) {
          const object2 = parseSerializableValue(firstArgument.arguments[0]);
          if (object2.kind !== "object") {
            return fallbackValue;
          }
          return map(object2);
        }
      }
    }
  }
  if (ts.isCallExpression(expression) && ts.isPropertyAccessExpression(expression.expression) && ts.isIdentifier(expression.expression.expression) && expression.expression.expression.text === "Promise" && ts.isIdentifier(expression.expression.name) && expression.arguments.length <= 1) {
    const firstArgument = expression.arguments[0];
    if (expression.expression.name.text === "resolve") {
      const value = firstArgument ? parseSerializableValue(firstArgument) : UNDEFINED;
      return promise({
        type: "resolve",
        value
      });
    }
    if (expression.expression.name.text === "reject") {
      if (!firstArgument) {
        return promise({
          type: "reject",
          message: null
        });
      } else if (ts.isNewExpression(firstArgument) && (firstArgument.arguments?.length || 0) <= 1) {
        const firstErrorArgument = firstArgument.arguments && firstArgument.arguments[0];
        if (firstErrorArgument) {
          if (ts.isStringLiteral(firstErrorArgument) || ts.isNoSubstitutionTemplateLiteral(firstErrorArgument)) {
            return promise({
              type: "reject",
              message: firstErrorArgument.text
            });
          }
        }
      }
    }
  }
  if (ts.isObjectLiteralExpression(expression)) {
    const entries = [];
    for (const property of expression.properties) {
      if (ts.isShorthandPropertyAssignment(property)) {
        entries.push({
          kind: "key",
          key: string(property.name.text),
          value: UNKNOWN
        });
      } else if (ts.isPropertyAssignment(property)) {
        const key = extractKeyFromPropertyName(property.name);
        if (key.kind === "unknown") {
          return fallbackValue;
        }
        entries.push({
          kind: "key",
          key,
          value: parseSerializableValue(property.initializer)
        });
      } else if (ts.isSpreadAssignment(property)) {
        entries.push({
          kind: "spread",
          value: property.name ? (
            // TODO: Property access off property.expression.
            UNKNOWN
          ) : parseSerializableValue(property.expression)
        });
      } else if (ts.isFunctionLike(property)) {
        return fallbackValue;
      } else {
        throw assertNever(property);
      }
    }
    return object(entries);
  }
  if (ts.isArrowFunction(expression) || ts.isFunctionExpression(expression)) {
    return fn(formatExpression(expression.getText()));
  }
  if (ts.isJsxElement(expression)) {
    return node(
      expression.openingElement.tagName.getText(),
      object(extractPropertiesFromJsxElement(expression.openingElement)),
      extractChildrenFromJsxElement(expression)
    );
  }
  if (ts.isJsxSelfClosingElement(expression)) {
    return node(
      expression.tagName.getText(),
      object(extractPropertiesFromJsxElement(expression)),
      null
    );
  }
  if (ts.isJsxFragment(expression)) {
    return node("", EMPTY_OBJECT, extractChildrenFromJsxElement(expression));
  }
  return fallbackValue;
}
function extractPropertiesFromJsxElement(element) {
  const objectEntries = [];
  for (const property of element.attributes.properties) {
    if (ts.isJsxSpreadAttribute(property)) {
      objectEntries.push({
        kind: "spread",
        value: parseSerializableValue(property.expression)
      });
    } else {
      let value;
      if (!property.initializer) {
        value = TRUE;
      } else if (ts.isJsxExpression(property.initializer)) {
        value = property.initializer.expression ? parseSerializableValue(property.initializer.expression) : unknown("");
      } else {
        value = parseSerializableValue(property.initializer);
      }
      objectEntries.push({
        kind: "key",
        key: string(property.name.text),
        value
      });
    }
  }
  return objectEntries;
}
function extractChildrenFromJsxElement(element) {
  const children = [];
  for (const child of element.children) {
    if (ts.isJsxText(child)) {
      if (child.containsOnlyTriviaWhiteSpaces) {
        continue;
      }
      children.push(
        string(
          // Note: we don't want to trim pure whitespace strings (typically separators).
          child.text.trim().length === 0 ? child.text : child.text.trim()
        )
      );
    } else if (ts.isJsxExpression(child)) {
      if (child.expression) {
        children.push(parseSerializableValue(child.expression));
      } else {
        children.push(unknown(""));
      }
    } else {
      children.push(parseSerializableValue(child));
    }
  }
  return children;
}
function extractKeyFromPropertyName(propertyName) {
  if (ts.isIdentifier(propertyName) || ts.isStringLiteral(propertyName) || ts.isNumericLiteral(propertyName)) {
    return string(propertyName.text);
  } else if (ts.isComputedPropertyName(propertyName)) {
    return parseSerializableValue(propertyName.expression);
  }
  return UNKNOWN;
}

export { EMPTY_ARRAY, EMPTY_MAP, EMPTY_OBJECT, EMPTY_SET, FALSE, NULL, TRUE, UNDEFINED, UNKNOWN, array, boolean, fn, generateSerializableValue, map, node, number, object, parseSerializableValue, promise, serializableValueToJavaScript, set, string, unknown };
